import threading
from queue import Queue
from .sox_player import ISoXPlayer

class SoundThread(threading.Thread):
    def __init__(self, q: Queue, play: ISoXPlayer):
        threading.Thread.__init__(self)
        self.q = q
        self.play = play

    def run(self):
        while True:
            if self.q.empty():
                self.play.dialer()
                continue
            return



import os
from queue import Queue

from .app_logger import IAppLogger
from .device_state import DeviceState
from .dnsleaktest import IDnsLeakTest
from .sox_player import ISoXPlayer
from .sound_thread import SoundThread

class NetworkManagerDispatcherApp:
    def __init__(self, device, state, logger: IAppLogger, play: ISoXPlayer, dnsleaktest: IDnsLeakTest):
        self.device = device # the network device name wlp1s0
        self.state = state   # up,down,connectivity-change etc.
        self.logger = logger
        self.play = play
        self.dnsleaktest = dnsleaktest

    def down(self):
        self.play.dialTone()
        return DeviceState.DOWN

    def up(self):
        # Create a Queue for the soundThread to check at some frequency to determine when it should quit.
        soundQueue = Queue(1)

        dialerThread = SoundThread(soundQueue, self.play)
        stdout = None
        stderr = None
        
        try:
            dialerThread.start()
            stdout, stderr = self.dnsleaktest.request()
        except Exception as ex:
            self.logger.log_error(f"Exception dnsleaktest.request.\n {ex}")
        finally:
            # Notify the soundThread that it can quit processing.
            soundQueue.put('quit')

            # Wait for the soundThread to terminate.
            dialerThread.join()

            # If there was an error, log it and return.
            if stderr:
                err = stderr.decode("utf-8")
                self.logger.log_error(err)
                self.play.playSoundForTestResult(err)
                return DeviceState.NO_INTERNET
            
            out = stdout.decode("utf-8")
            self.logger.log_info(out)

            self.play.playSoundForTestResult(out)
            return DeviceState.UP      

    def connectivity_change(self):
        NM_APP_STATE = os.environ.get("NM_APP_STATE")
        NM_APP_PREV_EVENT = os.environ.get("NM_APP_PREV_EVENT")
        
        # If we connected to the router which was broadcasting succesfully but the router couldn't establish
        # a connection through the gateway to any external domain when we performed the DNSLEAK test then the
        # las state will be NO_INTERNET. Finally since we are in the connectivity-change handler, then go ahead
        # and attempt the DNSLEAK test again.
        if NM_APP_STATE is not None and NM_APP_PREV_EVENT is not None:
            if int(NM_APP_STATE) == DeviceState.NO_INTERNET and NM_APP_PREV_EVENT == "up":
                return self.up()

    def main(self):
        try:
            self.logger.log_state(self.state)
            NM_APP_STATE = None
            
            if self.state == 'connectivity-change':
                NM_APP_STATE = self.connectivity_change()
 
            if self.state == "up":
                NM_APP_STATE = self.up()

            elif self.state == "down":
                NM_APP_STATE = self.down()

            os.environ.setdefault("NM_APP_PREV_EVENT", self.state)
            os.environ.setdefault("NM_APP_STATE", NM_APP_STATE)
        except Exception as ex:
            self.logger.log_error(f"Exception oops.\n {ex}")
        finally:
            self.logger.close()



from abc import ABC, abstractclassmethod
from io import IOBase

class IAppLogger(ABC):
    @abstractclassmethod
    def log_state(self, state):
        pass

    @abstractclassmethod
    def log_info(self, data):
        pass

    @abstractclassmethod    
    def log_error(self, data):
        pass

    @abstractclassmethod
    def close(self):
        pass

class AppLogger(IAppLogger):
    def __init__(self, file: IOBase, start: str) -> None:
        self.file = file 
        self.logstart = start

    def log_state(self, state):
        print(f"{self.logstart} Connection state '{state}' dispatched.", file=self.file, flush=True)

    def log_info(self, data):
        print(data, file=self.file, flush=True)

    def log_error(self, data):
        print(f'Error: {data}', file=self.file, flush=True)

    def close(self):
        self.file.close()



import time
import subprocess
from abc import ABC, abstractclassmethod

class ISoXPlayer(ABC):
    @abstractclassmethod
    def bell(self):
        pass
    
    @abstractclassmethod
    def dialer(self):
        pass
    
    @abstractclassmethod
    def dialTone(self):
        pass
    
    @abstractclassmethod
    def leakTone(self):
        pass

    @abstractclassmethod
    def playSoundForTestResult(self, dns_leak_test_result: str):
        pass

class ISubprocessCaller(ABC):
    def call(*popenargs, timeout=None, **kwargs):
        pass

class SoXPlayer(ISoXPlayer):
    def __init__(self, sub: ISubprocessCaller) -> None:
        super().__init__()
        self.sub = sub
    def call(self, cmd: str, shell: bool):
        self.sub.call(cmd, shell=shell)

    def bell(self):        
        self.call('play -q -n synth 3 sin 960 vol 0.1 fade l 0 3 2.8 2>/dev/null', shell=True)
    
    def dialer(self):
        self.call("play -q -n synth 0.1 sine 440 sine 480 channels 1 repeat 20 2>/dev/null", shell=True)
        time.sleep(4)
    
    def dialTone(self):
        self.call('play -q -n synth 0.1 sine 350 sine 440 channels 1 repeat 20 2>/dev/null', shell=True)

    def leakTone(self):
        self.call('play -q -n -c1 synth sin %-0 sin %-1 fade h 0.1 1 0.1 2>/dev/null', shell=True)

    def playSoundForTestResult(self, dns_leak_test_result: str):
        if 'DNS is not leaking' in dns_leak_test_result:
            self.bell()     
        elif 'DNS may be leaking' in dns_leak_test_result: 
            self.leakTone() 
        else:
            self.dialTone()



from typing import Tuple
from subprocess import Popen, PIPE
from abc import ABC, abstractclassmethod

class IDnsLeakTest(ABC):
    @abstractclassmethod
    def request(self) -> Tuple[bytes, bytes]:
        raise NotImplementedError

class DnsLeakTest(IDnsLeakTest):
    def request(self) -> Tuple[bytes, bytes]:
        try:
            # Run the bash.ws.sh script to retreive the dnsleaktest results.
            dnsleaktestProcess = Popen( \
                ['/etc/NetworkManager/dispatcher.d/network-manager-dispatcher-app-scripts/dnsleak.sh'],\
                    stdout=PIPE, \
                    stderr=PIPE \
            )
            return dnsleaktestProcess.communicate()
        except TimeoutError:
            dnsleaktestProcess.kill()
            return dnsleaktestProcess.communicate()

